/**
 * @function renderMiddleware
 */
'use strict'

const co = require('co')

let normalizeArray = (values) => [].concat(values || [])
  .filter(Boolean)
  .reduce((a, b) => [].concat(a, b), [])

let intercept = (instance, handlers, ...args) => co(function * () {
  for (let handler of normalizeArray(handlers)) {
    if (typeof handler !== 'function') {
      throw new Error(`Handler must be an function: ${handler}`)
    }
    yield Promise.resolve(handler.call(instance, ...args))
  }
})

/** @lends renderMiddleware */
function renderMiddleware (config) {
  let { name, methods, before, after, compile } = config
  let module = Object.assign({}, methods)
  return co.wrap(function * renderMiddleware (ctx, next) {
    let instance = new Proxy({ name, ctx }, {
      get (scope, name) {
        if (name in module) {
          return module[ name ]
        }
        if (name in scope) {
          return scope[ name ]
        }
        return ctx[ name ]
      }
    })
    ctx.render = co.wrap(function * render (methodName, params) {
      let method = methods[ methodName ]
      if (!method) {
        throw new Error(`[rview] Method not found: ${methodName}`)
      }
      let returns
      yield intercept(instance, before, methodName, params)
      returns = yield Promise.resolve(method.call(instance, params))
      if (compile) {
        compile = [].concat(compile).filter(Boolean)
        for (let c of compile) {
          returns = yield Promise.resolve(c.call(instance, returns))
        }
      }
      yield intercept(instance, after, methodName, params, returns)

      return returns
    })
    yield next()
  })
}

module.exports = renderMiddleware
